@@ -1,7 +1,3 @@
-# pylint:  disable=too-many-lines
-# (Yes, it has a point!)
-
-
 __copyright__ = """
 Copyright (C) 2009-2013 Andreas Kloeckner
 Copyright (C) 2020 Matt Wala
@@ -36,19 +32,42 @@ import sys
 from functools import reduce, wraps
 from sys import intern
 from typing import (
-    Any, Callable, ClassVar, Dict, Generic, Hashable, Iterable, List, Mapping,
-    Optional, Set, Tuple, Type, TypeVar, Union)
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    ClassVar,
+    Collection,
+    Dict,
+    Generic,
+    Hashable,
+    Iterable,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
+    Set,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+)
 
 
-try:
-    from typing import Concatenate, SupportsIndex
-except ImportError:
-    from typing_extensions import Concatenate, SupportsIndex
+if TYPE_CHECKING:
+    # NOTE: mypy seems to be confused by the `try.. except` below when called with
+    #   python -m mypy --python-version 3.8 ...
+    # see https://github.com/python/mypy/issues/14220
+    from typing_extensions import Concatenate, ParamSpec, SupportsIndex
+else:
+    try:
+        from typing import Concatenate, SupportsIndex
+    except ImportError:
+        from typing_extensions import Concatenate, SupportsIndex
 
-try:
-    from typing import ParamSpec
-except ImportError:
-    from typing_extensions import ParamSpec  # type: ignore[assignment]
+    try:
+        from typing import ParamSpec
+    except ImportError:
+        from typing_extensions import ParamSpec  # type: ignore[assignment]
 
 
 # These are deprecated and will go away in 2022.
@@ -189,6 +208,18 @@ String utilities
 ----------------
 
 .. autofunction:: strtobool
+.. autofunction:: to_identifier
+
+Set-like functions for iterables
+--------------------------------
+
+These functions provide set-like operations on iterables. In contrast to
+Python's built-in set type, they maintain the internal order of elements.
+
+.. autofunction:: unique
+.. autofunction:: unique_difference
+.. autofunction:: unique_intersection
+.. autofunction:: unique_union
 
 Type Variables Used
 -------------------
@@ -408,7 +439,10 @@ class RecordWithoutPickling:
     __slots__: ClassVar[List[str]] = []
     fields: ClassVar[Set[str]]
 
-    def __init__(self, valuedict=None, exclude=None, **kwargs):
+    def __init__(self,
+                 valuedict: Optional[Mapping[str, Any]] = None,
+                 exclude: Optional[Sequence[str]] = None,
+                 **kwargs: Any) -> None:
         assert self.__class__ is not Record
 
         if exclude is None:
@@ -443,7 +477,7 @@ class RecordWithoutPickling:
         return "{}({})".format(
                 self.__class__.__name__,
                 ", ".join(f"{fld}={getattr(self, fld)!r}"
-                    for fld in self.__class__.fields
+                    for fld in sorted(self.__class__.fields)
                     if hasattr(self, fld)))
 
     def register_fields(self, new_fields):
@@ -500,9 +534,10 @@ class ImmutableRecordWithoutPickling(RecordWithoutPickling):
     def __hash__(self):
         # This attribute may vanish during pickling.
         if getattr(self, "_cached_hash", None) is None:
-            self._cached_hash = hash(
-                (type(self),) + tuple(getattr(self, field)
-                    for field in self.__class__.fields))
+            self._cached_hash = hash((
+                    type(self),
+                    *(getattr(self, field) for field in self.__class__.fields)
+                    ))
 
         return self._cached_hash
 
@@ -519,7 +554,8 @@ class Reference:
 
     def get(self):
         from warnings import warn
-        warn("Reference.get() is deprecated -- use ref.value instead")
+        warn("Reference.get() is deprecated -- use ref.value instead. "
+             "This will stop working in 2025.", stacklevel=2)
         return self.value
 
     def set(self, value):
@@ -557,7 +593,7 @@ class DependentDictionary:
 
     def __contains__(self, key):
         try:
-            self[key]  # pylint: disable=pointless-statement
+            self[key]
             return True
         except KeyError:
             return False
@@ -571,7 +607,7 @@ class DependentDictionary:
     def __setitem__(self, key, value):
         self._Dictionary[key] = value
 
-    def genuineKeys(self):  # noqa
+    def genuineKeys(self):  # noqa: N802
         return list(self._Dictionary.keys())
 
     def iteritems(self):
@@ -598,7 +634,7 @@ def one(iterable: Iterable[T]) -> T:
     try:
         v = next(it)
     except StopIteration:
-        raise ValueError("empty iterable passed to 'one()'")
+        raise ValueError("empty iterable passed to 'one()'") from None
 
     def no_more():
         try:
@@ -620,7 +656,7 @@ def is_single_valued(
     try:
         first_item = next(it)
     except StopIteration:
-        raise ValueError("empty iterable passed to 'single_valued()'")
+        raise ValueError("empty iterable passed to 'single_valued()'") from None
 
     for other_item in it:
         if not equality_pred(other_item, first_item):
@@ -647,7 +683,7 @@ def single_valued(
     try:
         first_item = next(it)
     except StopIteration:
-        raise ValueError("empty iterable passed to 'single_valued()'")
+        raise ValueError("empty iterable passed to 'single_valued()'") from None
 
     def others_same():
         for other_item in it:
@@ -680,7 +716,7 @@ def memoize(*args: F, **kwargs: Any) -> F:
     default_key_func: Optional[Callable[..., Any]]
 
     if use_kw:
-        def default_key_func(*inner_args, **inner_kwargs):  # noqa pylint:disable=function-redefined
+        def default_key_func(*inner_args, **inner_kwargs):
             return inner_args, frozenset(inner_kwargs.items())
     else:
         default_key_func = None
@@ -697,15 +733,15 @@ def memoize(*args: F, **kwargs: Any) -> F:
             def wrapper(*args, **kwargs):
                 key = key_func(*args, **kwargs)
                 try:
-                    return func._memoize_dic[key]  # noqa: E501 # pylint: disable=protected-access
+                    return func._memoize_dic[key]
                 except AttributeError:
                     # _memoize_dic doesn't exist yet.
                     result = func(*args, **kwargs)
-                    func._memoize_dic = {key: result}  # noqa: E501 # pylint: disable=protected-access
+                    func._memoize_dic = {key: result}
                     return result
                 except KeyError:
                     result = func(*args, **kwargs)
-                    func._memoize_dic[key] = result  # noqa: E501 # pylint: disable=protected-access
+                    func._memoize_dic[key] = result
                     return result
 
             from functools import update_wrapper
@@ -716,15 +752,15 @@ def memoize(*args: F, **kwargs: Any) -> F:
         def _decorator(func):
             def wrapper(*args):
                 try:
-                    return func._memoize_dic[args]  # noqa: E501 # pylint: disable=protected-access
+                    return func._memoize_dic[args]
                 except AttributeError:
                     # _memoize_dic doesn't exist yet.
                     result = func(*args)
-                    func._memoize_dic = {args: result}  # noqa: E501 # pylint:disable=protected-access
+                    func._memoize_dic = {args: result}
                     return result
                 except KeyError:
                     result = func(*args)
-                    func._memoize_dic[args] = result  # noqa: E501 # pylint: disable=protected-access
+                    func._memoize_dic[args] = result
                     return result
 
             from functools import update_wrapper
@@ -762,7 +798,7 @@ def memoize_on_first_arg(
 
     def wrapper(obj: T, *args: P.args, **kwargs: P.kwargs) -> R:
         if kwargs:
-            key = (_HasKwargs, frozenset(kwargs.items())) + args
+            key = (_HasKwargs, frozenset(kwargs.items()), *args)
         else:
             key = args
 
@@ -803,7 +839,7 @@ def memoize_method(
     .. versionchanged:: 2021.2
 
         Can memoize methods on classes that do not allow setting attributes
-        (e.g. by overwritting ``__setattr__``), e.g. frozen :mod:`dataclasses`.
+        (e.g. by overwriting ``__setattr__``), e.g. frozen :mod:`dataclasses`.
     """
 
     return memoize_on_first_arg(method,
@@ -882,13 +918,13 @@ class keyed_memoize_method(keyed_memoize_on_first_arg):  # noqa: N801
     .. versionchanged:: 2021.2
 
         Can memoize methods on classes that do not allow setting attributes
-        (e.g. by overwritting ``__setattr__``), e.g. frozen :mod:`dataclasses`.
+        (e.g. by overwriting ``__setattr__``), e.g. frozen :mod:`dataclasses`.
     """
     def _default_cache_dict_name(self, function):
         return intern(f"_memoize_dic_{function.__name__}")
 
 
-class memoize_in(Generic[P, R]):  # noqa
+class memoize_in:  # noqa: N801
     """Adds a cache to the function it decorates. The cache is attached
     to *container* and must be uniquely specified by *identifier* (i.e.
     all functions using the same *container* and *identifier* will be using
@@ -908,7 +944,7 @@ class memoize_in(Generic[P, R]):  # noqa
     .. versionchanged:: 2021.2.1
 
         Can now use instances of classes as *container* that do not allow
-        setting attributes (e.g. by overwritting ``__setattr__``),
+        setting attributes (e.g. by overwriting ``__setattr__``),
         e.g. frozen :mod:`dataclasses`.
     """
 
@@ -934,13 +970,10 @@ class memoize_in(Generic[P, R]):  # noqa
                 self.cache_dict[args] = result
                 return result
 
-        # NOTE: mypy gets confused because it types `wraps` as
-        #   Callable[[VarArg(Any)], Any]
-        # which, for some reason, is not compatible with `F`
-        return new_inner                # type: ignore[return-value]
+        return new_inner
 
 
-class keyed_memoize_in(Generic[P, R]):  # noqa
+class keyed_memoize_in(Generic[P, R]):  # noqa: N801
     """Like :class:`memoize_in`, but additionally uses a function *key* to
     compute the key under which the function result is memoized.
 
@@ -1026,11 +1059,11 @@ def monkeypatch_class(_name, bases, namespace):
 # {{{ generic utilities
 
 def add_tuples(t1, t2):
-    return tuple([t1v + t2v for t1v, t2v in zip(t1, t2)])
+    return tuple(t1v + t2v for t1v, t2v in zip(t1, t2))
 
 
 def negate_tuple(t1):
-    return tuple([-t1v for t1v in t1])
+    return tuple(-t1v for t1v in t1)
 
 
 def shift(vec, dist):
@@ -1043,7 +1076,7 @@ def shift(vec, dist):
 
     result = vec[:]
 
-    N = len(vec)  # noqa
+    N = len(vec)  # noqa: N806
     dist = dist % N
 
     # modulo only returns positive distances!
@@ -1235,7 +1268,7 @@ def argmin2(iterable, return_value=False):
     try:
         current_argmin, current_min = next(it)
     except StopIteration:
-        raise ValueError("argmin of empty iterable")
+        raise ValueError("argmin of empty iterable") from None
 
     for arg, item in it:
         if item < current_min:
@@ -1253,7 +1286,7 @@ def argmax2(iterable, return_value=False):
     try:
         current_argmax, current_max = next(it)
     except StopIteration:
-        raise ValueError("argmax of empty iterable")
+        raise ValueError("argmax of empty iterable") from None
 
     for arg, item in it:
         if item > current_max:
@@ -1286,7 +1319,7 @@ def cartesian_product(*args):
     first = args[:-1]
     for prod in cartesian_product(*first):
         for i in args[-1]:
-            yield prod + (i,)
+            yield (*prod, i)
 
 
 def distinct_pairs(list1, list2):
@@ -1320,7 +1353,7 @@ def average(iterable):
         s = next(it)
         count = 1
     except StopIteration:
-        raise ValueError("empty average")
+        raise ValueError("empty average") from None
 
     for value in it:
         s = s + value
@@ -1400,7 +1433,7 @@ def indices_in_shape(shape):
         remainder = shape[1:]
         for i in range(0, shape[0]):
             for rest in indices_in_shape(remainder):
-                yield (i,)+rest
+                yield (i, *rest)
 
 
 def generate_nonnegative_integer_tuples_below(n, length=None, least=0):
@@ -1435,7 +1468,7 @@ def generate_decreasing_nonnegative_tuples_summing_to(
         yield ()
     elif length == 1:
         if n <= max_value:
-            #print "MX", n, max_value
+            # print "MX", n, max_value
             yield (n,)
         else:
             return
@@ -1444,10 +1477,10 @@ def generate_decreasing_nonnegative_tuples_summing_to(
             max_value = n
 
         for i in range(min_value, max_value+1):
-            #print "SIG", sig, i
+            # print "SIG", sig, i
             for remainder in generate_decreasing_nonnegative_tuples_summing_to(
                     n-i, length-1, min_value, i):
-                yield (i,) + remainder
+                yield (i, *remainder)
 
 
 def generate_nonnegative_integer_tuples_summing_to_at_most(n, length):
@@ -1462,7 +1495,7 @@ def generate_nonnegative_integer_tuples_summing_to_at_most(n, length):
         for i in range(n+1):
             for remainder in generate_nonnegative_integer_tuples_summing_to_at_most(
                     n-i, length-1):
-                yield remainder + (i,)
+                yield (*remainder, i)
 
 
 # backwards compatibility
@@ -1496,7 +1529,7 @@ def generate_permutations(original):
     else:
         for perm_ in generate_permutations(original[1:]):
             for i in range(len(perm_)+1):
-                #nb str[0:1] works in both string and list contexts
+                # nb str[0:1] works in both string and list contexts
                 yield perm_[:i] + original[0:1] + perm_[i:]
 
 
@@ -1519,68 +1552,6 @@ def enumerate_basic_directions(dimensions):
 # }}}
 
 
-# {{{ index mangling
-
-def get_read_from_map_from_permutation(original, permuted):
-    """With a permutation given by *original* and *permuted*,
-    generate a list *rfm* of indices such that
-    ``permuted[i] == original[rfm[i]]``.
-
-    Requires that the permutation can be inferred from
-    *original* and *permuted*.
-
-    .. doctest ::
-
-        >>> for p1 in generate_permutations(list(range(5))):
-        ...     for p2 in generate_permutations(list(range(5))):
-        ...         rfm = get_read_from_map_from_permutation(p1, p2)
-        ...         p2a = [p1[rfm[i]] for i in range(len(p1))]
-        ...         assert p2 == p2a
-    """
-    from warnings import warn
-    warn("get_read_from_map_from_permutation is deprecated and will be "
-            "removed in 2019", DeprecationWarning, stacklevel=2)
-
-    assert len(original) == len(permuted)
-    where_in_original = {
-            original[i]: i for i in range(len(original))}
-    assert len(where_in_original) == len(original)
-    return tuple(where_in_original[pi] for pi in permuted)
-
-
-def get_write_to_map_from_permutation(original, permuted):
-    """With a permutation given by *original* and *permuted*,
-    generate a list *wtm* of indices such that
-    ``permuted[wtm[i]] == original[i]``.
-
-    Requires that the permutation can be inferred from
-    *original* and *permuted*.
-
-    .. doctest ::
-
-        >>> for p1 in generate_permutations(list(range(5))):
-        ...     for p2 in generate_permutations(list(range(5))):
-        ...         wtm = get_write_to_map_from_permutation(p1, p2)
-        ...         p2a = [0] * len(p2)
-        ...         for i, oi in enumerate(p1):
-        ...             p2a[wtm[i]] = oi
-        ...         assert p2 == p2a
-    """
-    from warnings import warn
-    warn("get_write_to_map_from_permutation is deprecated and will be "
-            "removed in 2019", DeprecationWarning, stacklevel=2)
-
-    assert len(original) == len(permuted)
-
-    where_in_permuted = {
-            permuted[i]: i for i in range(len(permuted))}
-
-    assert len(where_in_permuted) == len(permuted)
-    return tuple(where_in_permuted[oi] for oi in original)
-
-# }}}
-
-
 # {{{ graph algorithms
 
 from pytools.graph import a_star as a_star_moved
@@ -1643,7 +1614,7 @@ class Table:
                     f"tried to add a row with {len(row)} columns to "
                     f"a table with {self.ncolumns} columns")
 
-        self.rows.append(tuple([str(i) for i in row]))
+        self.rows.append(tuple(str(i) for i in row))
 
     def _get_alignments(self) -> Tuple[str, ...]:
         # NOTE: If not all alignments were specified, extend alignments with the
@@ -1653,9 +1624,9 @@ class Table:
                 )
 
     def _get_column_widths(self, rows) -> Tuple[int, ...]:
-        return tuple([
+        return tuple(
             max(len(row[i]) for row in rows) for i in range(self.ncolumns)
-            ])
+            )
 
     def __str__(self) -> str:
         """
@@ -1704,7 +1675,7 @@ class Table:
             :--|-------:
             10 | 20\|\|
 
-        """  # noqa: W605
+        """
         if not self.rows:
             return ""
 
@@ -1712,7 +1683,7 @@ class Table:
             # Pipe symbols ('|') must be replaced
             return cell.replace("|", "\\|")
 
-        rows = [tuple([escape(cell) for cell in row]) for row in self.rows]
+        rows = [tuple(escape(cell) for cell in row) for row in self.rows]
         alignments = self._get_alignments()
         col_widths = self._get_column_widths(rows)
 
@@ -1822,14 +1793,14 @@ def merge_tables(*tables: Table,
         if i == 0 or skip_columns is None:
             return row
         else:
-            return tuple([
+            return tuple(
                 entry for i, entry in enumerate(row) if i not in skip_columns
-                ])
+                )
 
-    alignments = sum([
+    alignments = sum((
         remove_columns(i, tbl._get_alignments())
         for i, tbl in enumerate(tables)
-        ], ())
+        ), ())
     result = Table(alignments=alignments)
 
     for i in range(tables[0].nrows):
@@ -1846,7 +1817,7 @@ def merge_tables(*tables: Table,
 
 # {{{ histogram formatting
 
-def string_histogram(  # pylint: disable=too-many-arguments,too-many-locals
+def string_histogram(
         iterable, min_value=None, max_value=None,
         bin_count=20, width=70, bin_starts=None, use_unicode=True):
     if bin_starts is None:
@@ -1862,9 +1833,9 @@ def string_histogram(  # pylint: disable=too-many-arguments,too-many-locals
 
     from bisect import bisect
     for value in iterable:
-        if max_value is not None and value > max_value or value < bin_starts[0]:
+        if (max_value is not None and value > max_value) or value < bin_starts[0]:
             from warnings import warn
-            warn("string_histogram: out-of-bounds value ignored")
+            warn("string_histogram: out-of-bounds value ignored", stacklevel=2)
         else:
             bin_nr = bisect(bin_starts, value)-1
             try:
@@ -2007,7 +1978,6 @@ class CPyUserInterface:
 
 class StderrToStdout:
     def __enter__(self):
-        # pylint: disable=attribute-defined-outside-init
         self.stderr_backup = sys.stderr
         sys.stderr = sys.stdout
 
@@ -2117,7 +2087,7 @@ def invoke_editor(s, filename="edit.txt", descr="the file"):
 
 # {{{ progress bars
 
-class ProgressBar:  # pylint: disable=too-many-instance-attributes
+class ProgressBar:
     """
     .. automethod:: draw
     .. automethod:: progress
@@ -2295,7 +2265,7 @@ class UniqueNameGenerator:
     .. automethod:: __call__
     """
     def __init__(self,
-            existing_names: Optional[Set[str]] = None,
+            existing_names: Optional[Collection[str]] = None,
             forced_prefix: str = ""):
         """
         Create a new :class:`UniqueNameGenerator`.
@@ -2307,7 +2277,7 @@ class UniqueNameGenerator:
         if existing_names is None:
             existing_names = set()
 
-        self.existing_names = existing_names.copy()
+        self.existing_names = set(existing_names)
         self.forced_prefix = forced_prefix
         self.prefix_to_counter: Dict[str, int] = {}
 
@@ -2369,7 +2339,7 @@ class UniqueNameGenerator:
 
         # }}}
 
-        for counter, var_name in generate_numbered_unique_names(based_on, counter):  # noqa: B020,B007,E501
+        for counter, var_name in generate_numbered_unique_names(based_on, counter):  # noqa: B020,B007
             if not self.is_name_conflicting(var_name):
                 break
 
@@ -2391,8 +2361,6 @@ class MinRecursionLimit:
         self.min_rec_limit = min_rec_limit
 
     def __enter__(self):
-        # pylint: disable=attribute-defined-outside-init
-
         self.prev_recursion_limit = sys.getrecursionlimit()
         new_limit = max(self.prev_recursion_limit, self.min_rec_limit)
         sys.setrecursionlimit(new_limit)
@@ -2443,7 +2411,7 @@ def download_from_web_if_not_present(url, local_name=None):
 
 # {{{ find git revisions
 
-def find_git_revision(tree_root):  # pylint: disable=too-many-locals
+def find_git_revision(tree_root):
     # Keep this routine self-contained so that it can be copy-pasted into
     # setup.py.
 
@@ -2481,7 +2449,7 @@ def find_git_revision(tree_root):  # pylint: disable=too-many-locals
     assert retcode is not None
     if retcode != 0:
         from warnings import warn
-        warn("unable to find git revision")
+        warn("unable to find git revision", stacklevel=1)
         return None
 
     return git_rev
@@ -2489,7 +2457,7 @@ def find_git_revision(tree_root):  # pylint: disable=too-many-locals
 
 def find_module_git_revision(module_file, n_levels_up):
     from os.path import dirname, join
-    tree_root = join(*([dirname(module_file)] + [".." * n_levels_up]))
+    tree_root = join(*([dirname(module_file), ".." * n_levels_up]))
 
     return find_git_revision(tree_root)
 
@@ -2583,7 +2551,7 @@ def _log_start_if_long(logger, sleep_duration, done_indicator,
                 sleep_duration)
 
 
-class ProcessLogger:  # pylint: disable=too-many-instance-attributes
+class ProcessLogger:
     """Logs the completion time of a (presumably) lengthy process to :mod:`logging`.
     Only uses a high log level if the process took perceptible time.
 
@@ -2595,7 +2563,7 @@ class ProcessLogger:  # pylint: disable=too-many-instance-attributes
 
     default_noisy_level = logging.INFO
 
-    def __init__(  # pylint: disable=too-many-arguments
+    def __init__(
             self, logger, description,
             silent_level=None, noisy_level=None, long_threshold_seconds=None):
         self.logger = logger
@@ -2654,7 +2622,7 @@ class ProcessLogger:  # pylint: disable=too-many-instance-attributes
 
         self.timer = ProcessTimer()
 
-    def done(  # pylint: disable=keyword-arg-before-vararg
+    def done(
             self, extra_msg=None, *extra_fmt_args):
         self.timer.done()
         self._done_indicator[0] = True
@@ -2760,7 +2728,8 @@ def natsorted(iterable, key=None, reverse=False):
     .. versionadded:: 2020.1
     """
     if key is None:
-        key = lambda x: x
+        def key(x):
+            return x
     return sorted(iterable, key=lambda y: natorder(key(y)), reverse=reverse)
 
 # }}}
@@ -2784,7 +2753,7 @@ def resolve_name(name):
     if sys.version_info >= (3, 9):
         # use the official version
         import pkgutil
-        return pkgutil.resolve_name(name)  # pylint: disable=no-member
+        return pkgutil.resolve_name(name)
 
     import importlib
 
@@ -2824,12 +2793,14 @@ def resolve_name(name):
 
 # {{{ unordered_hash
 
-def unordered_hash(hash_instance, iterable, hash_constructor=None):
+def unordered_hash(hash_instance: Any,
+                   iterable: Iterable[Any],
+                   hash_constructor: Optional[Callable[[], Any]] = None) -> Any:
     """Using a hash algorithm given by the parameter-less constructor
     *hash_constructor*, return a hash object whose internal state
     depends on the entries of *iterable*, but not their order. If *hash*
     is the instance returned by evaluating ``hash_constructor()``, then
-    the each entry *i* of the iterable must permit ``hash.upate(i)`` to
+    the each entry *i* of the iterable must permit ``hash.update(i)`` to
     succeed. An example of *hash_constructor* is ``hashlib.sha256``
     from :mod:`hashlib`.  ``hash.digest_size`` must also be defined.
     If *hash_constructor* is not provided, ``hash_instance.name`` is
@@ -2850,6 +2821,8 @@ def unordered_hash(hash_instance, iterable, hash_constructor=None):
         from functools import partial
         hash_constructor = partial(hashlib.new, hash_instance.name)
 
+    assert hash_constructor is not None
+
     h_int = 0
     for i in iterable:
         h_i = hash_constructor()
@@ -2929,7 +2902,7 @@ def sphere_sample_fibonacci(
         lattice.
 
     :returns: an :class:`~numpy.ndarray` of shape ``(3, npoints)``.
-    """     # noqa: E501
+    """
 
     import numpy as np
     if optimize is None:
@@ -2994,6 +2967,88 @@ def strtobool(val: Optional[str], default: Optional[bool] = None) -> bool:
 # }}}
 
 
+# {{{ to_identifier
+
+def to_identifier(s: str) -> str:
+    """Convert a string to a valid Python identifier, by removing
+    non-alphanumeric, non-underscore characters, and prepending an underscore
+    if the string starts with a numeric character.
+
+    :param s: The string to convert to an identifier.
+
+    :returns: The converted string.
+    """
+    if s.isidentifier():
+        return s
+
+    s = "".join(c for c in s if c.isalnum() or c == "_")
+
+    if len(s) == 0:
+        return "_"
+
+    if s[0].isdigit():
+        s = "_" + s
+
+    return s
+
+# }}}
+
+
+# {{{ unique
+
+def unique(seq: Iterable[T]) -> Collection[T]:
+    """Return unique elements in *seq*, removing all duplicates. The internal
+    order of the elements is preserved. See also
+    :func:`itertools.groupby` (which removes consecutive duplicates)."""
+    return dict.fromkeys(seq)
+
+
+def unique_difference(*args: Iterable[T]) -> Collection[T]:
+    r"""Return unique elements that are in the first iterable in *\*args* but not
+    in any of the others. The internal order of the elements is preserved."""
+    if not args:
+        return []
+
+    res = dict.fromkeys(args[0])
+    for seq in args[1:]:
+        for item in seq:
+            if item in res:
+                del res[item]
+
+    return res
+
+
+def unique_intersection(*args: Iterable[T]) -> Collection[T]:
+    r"""Return unique elements that are common to all iterables in *\*args*.
+    The internal order of the elements is preserved."""
+    if not args:
+        return []
+
+    res = dict.fromkeys(args[0])
+    for seq in args[1:]:
+        seq = set(seq)
+        res = {item: None for item in res if item in seq}
+
+    return res
+
+
+def unique_union(*args: Iterable[T]) -> Collection[T]:
+    r"""Return unique elements that are in any iterable in *\*args*.
+    The internal order of the elements is preserved."""
+    if not args:
+        return []
+
+    res: Dict[T, None] = {}
+    for seq in args:
+        for item in seq:
+            if item not in res:
+                res[item] = None
+
+    return res
+
+# }}}
+
+
 def _test():
     import doctest
     doctest.testmod()
