@@ -1,9 +1,11 @@
 from __future__ import annotations
 
+
 __copyright__ = """
 Copyright (C) 2009-2013 Andreas Kloeckner
 Copyright (C) 2020 Matt Wala
 Copyright (C) 2020 James Stevens
+Copyright (C) 2024 Addison Alvey-Blanco
 """
 
 __license__ = """
@@ -46,6 +48,8 @@ Graph Algorithms
 .. autofunction:: as_graphviz_dot
 .. autofunction:: validate_graph
 .. autofunction:: is_connected
+.. autofunction:: undirected_graph_from_edges
+.. autofunction:: get_reachable_nodes
 
 Type Variables Used
 -------------------
@@ -63,19 +67,43 @@ Type Variables Used
     is included as a key in the graph.
 """
 
-from typing import (Collection, Mapping, List, Optional, Any,
-                    Callable, Set, MutableSet, Dict, Iterator, Tuple,
-                    Hashable, TypeVar)
+from dataclasses import dataclass
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Collection,
+    Dict,
+    FrozenSet,
+    Generic,
+    Hashable,
+    Iterable,
+    Iterator,
+    List,
+    Mapping,
+    MutableSet,
+    Optional,
+    Protocol,
+    Set,
+    Tuple,
+    TypeVar,
+)
 
-try:
-    from typing import TypeAlias
-except ImportError:
+
+if TYPE_CHECKING:
+    # NOTE: mypy seems to be confused by the `try.. except` below when called with
+    #   python -m mypy --python-version 3.8 ...
+    # see https://github.com/python/mypy/issues/14220
     from typing_extensions import TypeAlias
+else:
+    try:
+        from typing import TypeAlias
+    except ImportError:
+        from typing_extensions import TypeAlias
 
 
 NodeT = TypeVar("NodeT", bound=Hashable)
 
-
 GraphT: TypeAlias[NodeT] = Mapping[NodeT, Collection[NodeT]]
 
 
@@ -103,7 +131,7 @@ def reverse_graph(graph: GraphT[NodeT]) -> GraphT[NodeT]:
 
 # {{{ a_star
 
-def a_star(  # pylint: disable=too-many-locals
+def a_star(
         initial_state: NodeT, goal_state: NodeT, neighbor_map: GraphT[NodeT],
         estimate_remaining_cost: Optional[Callable[[NodeT], float]] = None,
         get_step_cost: Callable[[Any, NodeT], float] = lambda x, y: 1
@@ -115,7 +143,6 @@ def a_star(  # pylint: disable=too-many-locals
     from heapq import heappop, heappush
 
     if estimate_remaining_cost is None:
-        # pylint: disable=function-redefined
         def estimate_remaining_cost(x: NodeT) -> float:
             if x != goal_state:
                 return 1
@@ -123,7 +150,7 @@ def a_star(  # pylint: disable=too-many-locals
                 return 0
 
     class AStarNode:
-        __slots__ = ["state", "parent", "path_cost"]
+        __slots__ = ["parent", "path_cost", "state"]
 
         def __init__(self, state: NodeT, parent: Any, path_cost: float) -> None:
             self.state = state
@@ -241,7 +268,13 @@ class CycleError(Exception):
         self.node = node
 
 
-class HeapEntry:
+class _SupportsLT(Protocol):
+    def __lt__(self, other: object) -> bool:
+        ...
+
+
+@dataclass(frozen=True)
+class _HeapEntry(Generic[NodeT]):
     """
     Helper class to compare associated keys while comparing the elements in
     heap operations.
@@ -249,11 +282,10 @@ class HeapEntry:
     Only needs to define :func:`pytools.graph.__lt__` according to
     <https://github.com/python/cpython/blob/8d21aa21f2cbc6d50aab3f420bb23be1d081dac4/Lib/heapq.py#L135-L138>.
     """
-    def __init__(self, node: NodeT, key: Any) -> None:
-        self.node = node
-        self.key = key
+    node: NodeT
+    key: _SupportsLT
 
-    def __lt__(self, other: "HeapEntry") -> bool:
+    def __lt__(self, other: _HeapEntry[NodeT]) -> bool:
         return self.key < other.key
 
 
@@ -285,7 +317,7 @@ def compute_topological_order(graph: GraphT[NodeT],
 
     # {{{ compute nodes_to_num_predecessors
 
-    nodes_to_num_predecessors = {node: 0 for node in graph}
+    nodes_to_num_predecessors = dict.fromkeys(graph, 0)
 
     for node in graph:
         for child in graph[node]:
@@ -299,7 +331,7 @@ def compute_topological_order(graph: GraphT[NodeT],
     # heap: list of instances of HeapEntry(n) where 'n' is a node in
     # 'graph' with no predecessor. Nodes with no predecessors are the
     # schedulable candidates.
-    heap = [HeapEntry(n, keyfunc(n))
+    heap = [_HeapEntry(n, keyfunc(n))
             for n, num_preds in nodes_to_num_predecessors.items()
             if num_preds == 0]
     heapify(heap)
@@ -314,7 +346,7 @@ def compute_topological_order(graph: GraphT[NodeT],
         for child in graph.get(node_to_be_scheduled, ()):
             nodes_to_num_predecessors[child] -= 1
             if nodes_to_num_predecessors[child] == 0:
-                heappush(heap, HeapEntry(child, keyfunc(child)))
+                heappush(heap, _HeapEntry(child, keyfunc(child)))
 
     if len(order) != total_num_nodes:
         # any node which has a predecessor left is a part of a cycle
@@ -435,10 +467,12 @@ def as_graphviz_dot(graph: GraphT[NodeT],
     from pytools.graphviz import dot_escape
 
     if node_labels is None:
-        node_labels = lambda x: str(x)
+        def node_labels(x: NodeT) -> str:
+            return str(x)
 
     if edge_labels is None:
-        edge_labels = lambda x, y: ""
+        def edge_labels(x: NodeT, y: NodeT) -> str:
+            return ""
 
     node_to_id = {}
 
@@ -463,7 +497,7 @@ def as_graphviz_dot(graph: GraphT[NodeT],
          for (node, targets) in graph.items()
          for t in targets])
 
-    return f"digraph mygraph {{\n{ content }\n}}\n"
+    return f"digraph mygraph {{\n{content}\n}}\n"
 
 # }}}
 
@@ -487,7 +521,7 @@ def validate_graph(graph: GraphT[NodeT]) -> None:
 # }}}
 
 
-# {{{
+# {{{ is_connected
 
 def is_connected(graph: GraphT[NodeT]) -> bool:
     """
@@ -518,5 +552,52 @@ def is_connected(graph: GraphT[NodeT]) -> bool:
 
     return visited == graph.keys()
 
+# }}}
+
+
+def undirected_graph_from_edges(
+            edges: Iterable[Tuple[NodeT, NodeT]],
+        ) -> GraphT[NodeT]:
+    """
+    Constructs an undirected graph using *edges*.
+
+    :arg edges: An :class:`Iterable` of pairs of related :class:`NodeT` s.
+
+    :returns: A :class:`GraphT` that is the undirected graph.
+    """
+    undirected_graph: Dict[NodeT, Set[NodeT]] = {}
+
+    for lhs, rhs in edges:
+        if lhs == rhs:
+            raise TypeError("Found loop in edges,"
+                            f" LHS, RHS = {lhs}")
+
+        undirected_graph.setdefault(lhs, set()).add(rhs)
+        undirected_graph.setdefault(rhs, set()).add(lhs)
+
+    return undirected_graph
+
+
+def get_reachable_nodes(
+        undirected_graph: GraphT[NodeT],
+        source_node: NodeT) -> FrozenSet[NodeT]:
+    """
+    Returns a :class:`frozenset` of all nodes in *undirected_graph* that are
+    reachable from *source_node*.
+    """
+    nodes_visited: Set[NodeT] = set()
+    nodes_to_visit = {source_node}
+
+    while nodes_to_visit:
+        current_node = nodes_to_visit.pop()
+        nodes_visited.add(current_node)
+
+        neighbors = undirected_graph[current_node]
+        nodes_to_visit.update({node
+                               for node in neighbors
+                               if node not in nodes_visited})
+
+    return frozenset(nodes_visited)
+
 
 # vim: foldmethod=marker
