@@ -1,6 +1,6 @@
-"""OpenMPI, once intialized, prohibits forking. This helper module
+"""OpenMPI, once initialized, prohibits forking. This helper module
 allows the forking of *one* helper child process before OpenMPI
-initializaton that can do the forking for the fork-challenged
+initialization that can do the forking for the fork-challenged
 parent process.
 
 Since none of this is MPI-specific, it got parked in pytools.
@@ -23,7 +23,8 @@ class DirectForker:
         try:
             return spcall(cmdline, cwd=cwd)
         except OSError as e:
-            raise ExecError("error invoking '{}': {}".format(" ".join(cmdline), e))
+            raise ExecError(
+                    "error invoking '{}': {}".format(" ".join(cmdline), e)) from e
 
     def call_async(self, cmdline, cwd=None):
         from subprocess import Popen
@@ -36,11 +37,12 @@ class DirectForker:
 
             return self.count
         except OSError as e:
-            raise ExecError("error invoking '{}': {}".format(" ".join(cmdline), e))
+            raise ExecError(
+                "error invoking '{}': {}".format(" ".join(cmdline), e)) from e
 
     @staticmethod
     def call_capture_output(cmdline, cwd=None, error_on_nonzero=True):
-        from subprocess import Popen, PIPE
+        from subprocess import PIPE, Popen
 
         try:
             popen = Popen(cmdline, cwd=cwd, stdin=PIPE, stdout=PIPE,
@@ -55,7 +57,8 @@ class DirectForker:
 
             return popen.returncode, stdout_data, stderr_data
         except OSError as e:
-            raise ExecError("error invoking '{}': {}".format(" ".join(cmdline), e))
+            raise ExecError(
+                    "error invoking '{}': {}".format(" ".join(cmdline), e)) from e
 
     def wait(self, aid):
         proc = self.apids.pop(aid)
@@ -73,8 +76,8 @@ class DirectForker:
 
 
 def _send_packet(sock, data):
-    from struct import pack
     from pickle import dumps
+    from struct import pack
 
     packet = dumps(data)
 
@@ -138,7 +141,7 @@ def _fork_server(sock):
         sock.close()
 
     import os
-    os._exit(0)  # pylint:disable=protected-access
+    os._exit(0)
 
 
 class IndirectForker:
@@ -188,7 +191,7 @@ forker = DirectForker()
 
 
 def enable_prefork():
-    global forker  # pylint:disable=global-statement
+    global forker
 
     if isinstance(forker, IndirectForker):
         return
